import numpy as np
import xmltodict

from pathlib import Path

BASE = Path(__file__).parent


class CG2:
    def __init__(self):
        self.Facility = "ORNL"
        self.Instrument = "CG2"
        # Filename
        self.filename_numeric_length = 4
        self.filename_lead_string = "xxxxxxxxx"
        self.filename_tail_string = []
        self.filename_extension_string = [".xml"]
        self.filename_data_loader = "raw_read_ornl_sans"
        # Misc
        self.guide_size = [40e-3, 40e-3]  # [x,y]
        self.attenuators = [
            1,
            3.605,
            43.85,
            553.544,
            3456,
            22973.37,
            69043,
        ]  # Attenautors 0, 1, 2, 3 etc.
        # Detector(s)
        self.detectors = 1
        # Main Panel
        self.detector1_type = "tube"
        self.detector1_view_position = "centre"
        self.detector1_name = "Rear"
        self.detector1_nominal_det_translation = [0, 0]
        self.detector1_dead_time = 0.6 * 10e-6
        self.detector1_pixels = [192, 256]
        self.detector1_pixel_size = [5.5, 3.94]  # mm [x y]
        self.detector1_nominal_beam_centre = [96.5, 128.5]  # [x,y]
        self.detector1_imask_file = "ornl_cg2_msk.msk"
        self.detector1_mask = np.empty((256, 192)) * np.nan
        self.detector1_relative_efficiency = 1
        self.detector1_efficiency_file = ""
        self.detector1_efficiency_data = np.ones((256, 192))
        self.detector1_efficiency_error = np.zeros((256, 192))

    def read_file(self, dir, path):
        self.path = dir + "/" + path
        with open(self.path) as fd:
            root = xmltodict.parse(fd.read(), process_namespaces=True)
        # print(root['SPICErack'].keys())
        self.SPICE_version = self.read_value(root["SPICErack"], "@SPICE_version")
        self.read_parameters(root["SPICErack"])

    def read_parameters(self, root):
        # Header
        self.Instrument = self.read_value(root["Header"], "Instrument")
        temp = self.read_value(root["Header"], "Start_Time")
        if temp is not None:
            self.Start_Date = temp[0:10]
            self.Start_Time = temp[-8:]
        temp = self.read_value(root["Header"], "End_Time")
        if temp is not None:
            self.End_Date = temp[0:10]
            self.End_Time = temp[-8:]
        self.Reactor_Power = self.read_typed_value(root["Header"], "Reactor_Power")
        self.Experiment_Title = self.read_value(root["Header"], "Experiment_Title")
        self.Experiment_number = self.read_typed_value(
            root["Header"], "Experiment_number"
        )
        self.IPTS_number = self.read_value(root["Header"], "IPTS_number")
        self.Cycle_Number = self.read_value(root["Header"], "Cycle_Number")
        self.Command = self.read_value(root["Header"], "Command")
        self.Users = self.read_value(root["Header"], "Users")
        self.Local_Contact = self.read_value(root["Header"], "Local_Contact")
        self.Scan_Number = self.read_typed_value(root["Header"], "Scan_Number")
        self.Scan_Point_Number = self.read_typed_value(
            root["Header"], "Scan_Point_Number"
        )
        self.Scan_Title = self.read_value(root["Header"], "Scan_Title")
        self.Sample_Name = self.read_value(root["Header"], "Sample_Name")
        self.Sample_ID = self.read_value(root["Header"], "Sample_ID")
        self.Sample_Changer = self.read_value(root["Header"], "Sample_Changer")
        self.Sample_Changer_Position = self.read_typed_value(
            root["Header"], "Sample_Changer_Position"
        )
        self.Sample_Thickness = self.read_typed_value(
            root["Header"], "Sample_Thickness"
        )
        self.Sample_CountRate = self.read_typed_value(
            root["Header"], "Sample_CountRate"
        )
        self.Transmission = self.read_value(root["Header"], "Transmission")
        self.Scan_Type = self.read_value(root["Header"], "Scan_Type")
        self.Sample_Type = self.read_value(root["Header"], "Sample_Type")
        self.Number_of_X_Pixels = self.read_typed_value(
            root["Header"], "Number_of_X_Pixels"
        )
        self.Number_of_Y_Pixels = self.read_typed_value(
            root["Header"], "Number_of_Y_Pixels"
        )
        self.west_wing_number_of_x_Pixels = self.read_typed_value(
            root["Header"], "west_wing_number_of_x_Pixels"
        )
        self.west_wing_number_of_y_Pixels = self.read_typed_value(
            root["Header"], "west_wing_number_of_y_Pixels"
        )
        self.west_wing_tank_center_internal_offset_mm = self.read_typed_value(
            root["Header"], "west_wing_tank_center_internal_offset_mm"
        )
        self.west_wing_det_radius_m = self.read_typed_value(
            root["Header"], "west_wing_det_radius_m"
        )
        self.west_wing_rot_offset_deg = self.read_typed_value(
            root["Header"], "west_wing_rot_offset_deg"
        )
        self.det_axis_offset_mm = self.read_typed_value(
            root["Header"], "det_axis_offset_mm"
        )
        self.x_mm_per_pixel = self.read_typed_value(root["Header"], "x_mm_per_pixel")
        self.y_mm_per_pixel = self.read_typed_value(root["Header"], "y_mm_per_pixel")
        self.absolute_intensity_constant = self.read_typed_value(
            root["Header"], "absolute_intensity_constant"
        )
        self.source_aperture_size = self.read_typed_value(
            root["Header"], "source_aperture_size"
        )
        self.sample_aperture_name = self.read_value(
            root["Header"], "sample_aperture_name"
        )
        self.sample_aperture_description = self.read_value(
            root["Header"], "sample_aperture_description"
        )
        self.sample_aperture_size = self.read_typed_value(
            root["Header"], "sample_aperture_size"
        )
        self.beamtrap_name = self.read_value(root["Header"], "beamtrap_name")
        self.beamtrap_diameter = self.read_typed_value(
            root["Header"], "beamtrap_diameter"
        )
        self.beamtrap_is_blocking = self.read_value(
            root["Header"], "beamtrap_is_blocking"
        )
        self.source_distance = self.read_typed_value(root["Header"], "source_distance")
        self.sample_to_flange = self.read_typed_value(
            root["Header"], "sample_to_flange"
        )
        self.sample_aperture_to_flange = self.read_typed_value(
            root["Header"], "sample_aperture_to_flange"
        )
        self.tank_internal_offset = self.read_typed_value(
            root["Header"], "tank_internal_offset"
        )
        self.wavelength_formula = self.read_value(root["Header"], "wavelength_formula")
        self.wavelength = self.read_typed_value(root["Header"], "wavelength")
        self.wavelength_spread = self.read_typed_value(
            root["Header"], "wavelength_spread"
        )
        self.source_distance = self.read_typed_value(root["Header"], "source_distance")
        self.Ref_Scan_Main_Flood = self.read_value(
            root["Header"], "Ref_Scan_Main_Flood"
        )
        self.Ref_Scan_Main_Beam_Center = self.read_value(
            root["Header"], "Ref_Scan_Main_Beam_Center"
        )
        self.Ref_Scan_Main_Dark_Current = self.read_value(
            root["Header"], "Ref_Scan_Main_Dark_Current"
        )
        self.Ref_Scan_Main_Bar = self.read_value(root["Header"], "Ref_Scan_Main_Bar")
        self.Ref_Scan_West_Wing_Flood = self.read_value(
            root["Header"], "Ref_Scan_West_Wing_Flood"
        )
        self.Ref_Scan_West_Wing_Dark_Current = self.read_value(
            root["Header"], "Ref_Scan_West_Wing_Dark_Current"
        )
        self.Ref_Scan_West_Wing_Bar = self.read_value(
            root["Header"], "Ref_Scan_West_Wing_Bar"
        )
        self.sans_spice_xml_format_version = self.read_typed_value(
            root["Header"], "sans_spice_xml_format_version"
        )
        self.Comment = self.read_value(root["Header"], "Comment")
        self.ImagePath = self.read_value(root["Header"], "ImagePath")
        # Sample Info
        self.Field = {}
        for key in root["sample_info"].keys():
            if root["sample_info"].get(key) is not None:
                for item in root["sample_info"][key].keys():
                    if key == "background" and item == "Field":
                        self.Field[
                            self.read_value(
                                root["sample_info"]["background"]["Field"], "@full_name"
                            )
                        ] = self.read_value(
                            root["sample_info"]["background"]["Field"], "#text"
                        )
        # Motor Positions
        self.highvoltagecmd = {}
        if root["Motor_Positions"]["highvoltagecmd"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["highvoltagecmd"], self.highvoltagecmd
            )
        self.selector_speed = {}
        if root["Motor_Positions"]["selector_speed"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["selector_speed"], self.selector_speed
            )
        self.selector_tilt = {}
        if root["Motor_Positions"]["selector_tilt"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["selector_tilt"], self.selector_tilt
            )
        self._lambda = {}
        if root["Motor_Positions"]["lambda"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["lambda"], self._lambda)
        self.dlambda = {}
        if root["Motor_Positions"]["dlambda"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["dlambda"], self.dlambda)
        self.sdd = {}
        if root["Motor_Positions"]["sdd"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["sdd"], self.sdd)
        self.coll_1 = {}
        if root["Motor_Positions"]["coll_1"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_1"], self.coll_1)
        self.coll_2 = {}
        if root["Motor_Positions"]["coll_2"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_2"], self.coll_2)
        self.coll_3 = {}
        if root["Motor_Positions"]["coll_3"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_3"], self.coll_3)
        self.coll_4 = {}
        if root["Motor_Positions"]["coll_4"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_4"], self.coll_4)
        self.coll_5 = {}
        if root["Motor_Positions"]["coll_5"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_5"], self.coll_5)
        self.coll_6 = {}
        if root["Motor_Positions"]["coll_6"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_6"], self.coll_6)
        self.coll_7 = {}
        if root["Motor_Positions"]["coll_7"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_7"], self.coll_7)
        self.coll_8 = {}
        if root["Motor_Positions"]["coll_8"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["coll_8"], self.coll_8)
        self.nguides = {}
        if root["Motor_Positions"]["nguides"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["nguides"], self.nguides)
        self.attenuator_pos = {}
        if root["Motor_Positions"]["attenuator_pos"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["attenuator_pos"], self.attenuator_pos
            )
        self.beam_trap_x = {}
        if root["Motor_Positions"]["beam_trap_x"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["beam_trap_x"], self.beam_trap_x
            )
        self.dcal = {}
        if root["Motor_Positions"]["dcal"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["dcal"], self.dcal)
        self.detector_trans = {}
        if root["Motor_Positions"]["detector_trans"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["detector_trans"], self.detector_trans
            )
        self.flange_det_dist = {}
        if root["Motor_Positions"]["flange_det_dist"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["flange_det_dist"], self.flange_det_dist
            )
        self.sample_x = {}
        if root["Motor_Positions"]["sample_x"].keys() is not None:
            self.read_positions(root["Motor_Positions"]["sample_x"], self.sample_x)
        self.trap_y_101mm = {}
        if root["Motor_Positions"]["trap_y_101mm"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["trap_y_101mm"], self.trap_y_101mm
            )
        self.trap_y_25mm = {}
        if root["Motor_Positions"]["trap_y_25mm"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["trap_y_25mm"], self.trap_y_101mm
            )
        self.trap_y_50mm = {}
        if root["Motor_Positions"]["trap_y_50mm"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["trap_y_50mm"], self.trap_y_50mm
            )
        self.trap_y_76mm = {}
        if root["Motor_Positions"]["trap_y_76mm"].keys() is not None:
            self.read_positions(
                root["Motor_Positions"]["trap_y_76mm"], self.trap_y_76mm
            )
        # Parameter Positions
        self.hvsts1 = {}
        if root["Parameter_Positions"]["hvsts1"].keys() is not None:
            self.read_positions(root["Parameter_Positions"]["hvsts1"], self.hvsts1)
        self.hvsts2 = {}
        if root["Parameter_Positions"]["hvsts2"].keys() is not None:
            self.read_positions(root["Parameter_Positions"]["hvsts2"], self.hvsts2)
        self.hvsts3 = {}
        if root["Parameter_Positions"]["hvsts3"].keys() is not None:
            self.read_positions(root["Parameter_Positions"]["hvsts3"], self.hvsts3)
        self.vs_current = {}
        if root["Parameter_Positions"]["vs_current"].keys() is not None:
            self.read_positions(
                root["Parameter_Positions"]["vs_current"], self.vs_current
            )
        self.vs_speed_pv = {}
        if root["Parameter_Positions"]["vs_speed_pv"].keys() is not None:
            self.read_positions(
                root["Parameter_Positions"]["vs_speed_pv"], self.vs_speed_pv
            )
        self.beam_on = {}
        if root["Parameter_Positions"]["beam_on"].keys() is not None:
            self.read_positions(root["Parameter_Positions"]["beam_on"], self.beam_on)
        # Counters
        self.time = {}
        if root["Counters"]["time"].keys() is not None:
            self.read_counters(root["Counters"]["time"], self.time)
        self.detector = {}
        if root["Counters"]["detector"].keys() is not None:
            self.read_counters(root["Counters"]["detector"], self.detector)
        self.monitor = {}
        if root["Counters"]["monitor"].keys() is not None:
            self.read_counters(root["Counters"]["monitor"], self.monitor)
        # Data
        self.data_description = self.read_value(root["Data"], "@description")
        temp = self.read_value(root["Data"]["Detector"], "@type")
        if temp[0:3] == "INT":
            data = np.fromstring(
                self.read_value(root["Data"]["Detector"], "#text"),
                dtype=np.intc,
                sep=" ",
            )
        else:
            data = np.fromstring(
                self.read_value(root["Data"]["Detector"], "#text"),
                dtype=np.single,
                sep=" ",
            )
        det = data.reshape(self.Number_of_X_Pixels, self.Number_of_Y_Pixels)
        self.det1_data = det.transpose()
        self.nx1, self.ny1 = self.det1_data.shape
        return

    def read_value(self, root, item, default=None):
        value = root.get(item)
        if value is None:
            # print(item, default)
            return default
        else:
            # print(item, value)
            return value

    def read_typed_value(self, root, item, default=None):
        value = root.get(item)
        if value is None:
            # print(item, default)
            return default
        else:
            typed = self.read_value(root[item], "@type")
            if typed == "FLOAT32":
                value = float(self.read_value(root[item], "#text"))
                # print(item, "FLOAT32", value)
                return value
            elif typed == "INT32":
                value = int(self.read_value(root[item], "#text"))
                # print(item, "INT32", value)
                return value
            else:
                # print(item, default)
                return default

    def read_positions(self, root, dict):
        dict["pos"] = self.read_value(root, "@pos")
        dict["units"] = self.read_value(root, "@units")
        dict["description"] = self.read_value(root, "@description")
        dict["type"] = self.read_value(root, "@type")
        dict["value"] = self.read_value(root, "#text")
        # print(dict)

    def read_counters(self, root, dict):
        dict["units"] = self.read_value(root, "@units")
        dict["description"] = self.read_value(root, "@description")
        dict["type"] = self.read_value(root, "@type")
        dict["value"] = self.read_value(root, "#text")
        # print(dict)



import scipy.io as sio


def read_mask_mat_file(filename):
    mat_contents = sio.loadmat(filename)
    return mat_contents["mask_data"]


def read_efficiency_mat_file_data(filename):
    mat_contents = sio.loadmat(filename)
    return mat_contents["eff_data"]


def read_efficiency_mat_file_error(filename):
    mat_contents = sio.loadmat(filename)
    return mat_contents["eff_err_data"]



from trame.widgets import html, vuetify


def create_transmission_content():
    html.Div("<strong>Samples</strong>")
    vuetify.VDivider(classes="mb-2")
    sample_files()
    html.Div("<strong>Empty Cell</strong>")
    vuetify.VDivider(classes="mb-2")
    empty_cell_files()
    html.Div("<strong>Direct Beam</strong>")
    vuetify.VDivider(classes="mb-2")
    direct_beam_files()


def sample_files():
    with vuetify.VRow(classes="pa-0 pt-2", dense=True):
        with vuetify.VCombobox(
            classes="mt=2",
            v_model=("transmission_samples_28m", []),
            items=("files",),
            label="28 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    close=True,
                    click="select",
                    click_close="transmission_samples_28m.splice(transmission_samples_28m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")
    with vuetify.VRow(classes="pa-0", dense=True):
        with vuetify.VCombobox(
            classes="mt=4",
            v_model=("transmission_samples_8m", []),
            items=("files",),
            label="8 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    close=True,
                    click="select",
                    click_close="transmission_samples_8m.splice(transmission_samples_8m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")
    with vuetify.VRow(classes="pa-0 pb-2", dense=True):
        with vuetify.VCombobox(
            v_model=("transmission_samples_2m", []),
            items=("files",),
            label="2 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    dense=True,
                    close=True,
                    click="select",
                    click_close="transmission_samples_2m.splice(transmission_samples_2m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")


def empty_cell_files():
    with vuetify.VRow(classes="pa-0 pt-2", dense=True):
        with vuetify.VCombobox(
            classes="mt=2",
            v_model=("transmission_empty_cell_28m", []),
            items=("files",),
            label="28 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    close=True,
                    click="select",
                    click_close="transmission_empty_cell_28m.splice(transmission_empty_cell_28m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")
    with vuetify.VRow(classes="pa-0", dense=True):
        with vuetify.VCombobox(
            classes="mt=4",
            v_model=("transmission_empty_cell_8m", []),
            items=("files",),
            label="8 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    close=True,
                    click="select",
                    click_close="transmission_empty_cell_8m.splice(transmission_empty_cell_8m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")
    with vuetify.VRow(classes="pa-0 pb-2", dense=True):
        with vuetify.VCombobox(
            v_model=("transmission_empty_cell_2m", []),
            items=("files",),
            label="2 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    dense=True,
                    close=True,
                    click="select",
                    click_close="transmission_empty_cell_2m.splice(transmission_empty_cell_2m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")


def direct_beam_files():
    with vuetify.VRow(classes="pa-0 pt-2", dense=True):
        with vuetify.VCombobox(
            classes="mt=2",
            v_model=("transmission_direct_beam_28m", []),
            items=("files",),
            label="28 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    close=True,
                    click="select",
                    click_close="transmission_direct_beam_28m.splice(transmission_direct_beam_28m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")
    with vuetify.VRow(classes="pa-0", dense=True):
        with vuetify.VCombobox(
            classes="mt=4",
            v_model=("transmission_direct_beam_8m", []),
            items=("files",),
            label="8 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    close=True,
                    click="select",
                    click_close="transmission_direct_beam_8m.splice(transmission_direct_beam_8m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")
    with vuetify.VRow(classes="pa-0 pb-2", dense=True):
        with vuetify.VCombobox(
            v_model=("transmission_direct_beam_2m", []),
            items=("files",),
            label="2 m",
            dense=True,
            hide_details=True,
            hide_no_data=True,
            hide_selected=True,
            chips=True,
            clearable=True,
            multiple=True,
            density="comfortable",
            outlined=True,
        ):
            with vuetify.Template(
                v_slot_selection="{ attrs, item, select, selected }",
            ):
                with vuetify.VChip(
                    v_bind="attrs",
                    input_value="selected",
                    dense=True,
                    close=True,
                    click="select",
                    click_close="transmission_direct_beam_2m.splice(transmission_direct_beam_2m.indexOf(item), 1);",
                ):
                    html.Div("<strong>{{ item }}</strong>")



from trame.widgets import html, vuetify
from .ui_scattering import create_scattering_content
from .ui_transmission import create_transmission_content

SUPPORTED_DEVICES = ["D11+", "CG2"]


class AbstractCard(vuetify.VCard):
    def __init__(self, key, icon, title):
        super().__init__(
            # classes="mb-3",
            tile=True,
            v_show=f"panel_visible.includes('{key}')",
        )
        with self:
            with vuetify.VCardTitle(
                classes="grey lighten-1 py-0 px-1 grey--text text--darken-3",
                style="user-select: none;",
                hide_details=True,
                dense=True,
            ) as self._title:
                vuetify.VIcon(icon, classes="mr-1", color="grey darken-3")
                html.Div(title)

            self._content = vuetify.VCardText(classes="pa-1")


class DeviceSelector(AbstractCard):
    def __init__(self):
        super().__init__("device", "mdi-microscope", "Device")
        with self._content:
            with vuetify.VRow(classes="px-0 py-1", dense=True, hide_details=True):
                with vuetify.VCol(cols="6"):
                    vuetify.VSelect(
                        v_model=("selectedDevice", None),
                        items=("devices", SUPPORTED_DEVICES),
                        label="Device",
                        dense=True,
                        hide_details=True,
                        outlined=True,
                    )
                with vuetify.VCol(cols="6"):
                    vuetify.VSelect(
                        v_model=("device_active_data", None),
                        items=("device_data", ["", "mask", "efficiency", "error"]),
                        label="Parameters",
                        dense=True,
                        hide_details=True,
                        outlined=True,
                    )


class DirectoryDialog(vuetify.VDialog):
    def __init__(self, select_directory_fn):
        super().__init__(
            v_model=("directory_dialog", False),
            persistent=True,
        )
        with self:
            with vuetify.Template(
                v_slot_activator="{ on, attrs }",
            ):
                with vuetify.VBtn(
                    small=True,
                    color="primary",
                    dark=True,
                    icon=True,
                    v_bind="attrs",
                    v_on="on",
                ):
                    vuetify.VIcon("mdi-folder")
            with vuetify.VCard(
                classes="ma-1 rounded elevation-8", style="min-width: 50vw;"
            ):
                with vuetify.VCardTitle(
                    classes="grey lighten-1 px-2 py-0 grey--text text--darken-3",
                    style="user-select: none;",
                    hide_details=True,
                    dense=True,
                ):
                    html.Div("Directory")
                with vuetify.VCardText(
                    classes="pa-1", style="height: 50vh; overflow: auto;"
                ):
                    vuetify.VTreeview(
                        classes="mb-1",
                        dense=True,
                        hide_details=True,
                        v_model=("dirtree", []),
                        items=("dirs", []),
                        activatable=True,
                        item_key="path",
                        open_on_click=True,
                        update_active=(
                            select_directory_fn,
                            "[$event]",
                        ),
                    )
                with vuetify.VCardActions(
                    classes="py-1",
                    hide_details=True,
                    dense=True,
                ):
                    vuetify.VSpacer()
                    with vuetify.VBtn(
                        icon=True,
                        dense=True,
                        small=True,
                        tile=True,
                        color="error",
                        click="directory_dialog = false",
                    ):
                        vuetify.VIcon("mdi-cancel")
                    with vuetify.VBtn(
                        icon=True,
                        dense=True,
                        small=True,
                        tile=True,
                        color="success",
                        click="directory_dialog = false",
                    ):
                        vuetify.VIcon("mdi-check")


class Directory(AbstractCard):
    def __init__(self, select_directory_fn):
        super().__init__("directory", "mdi-folder-eye-outline", "Directory")
        with self._content:
            with vuetify.VRow(classes="px-0 py-1", align_self="center"):
                with vuetify.VCol(cols="1", align_self="center"):
                    DirectoryDialog(select_directory_fn)
                with vuetify.VCol(cols="11"):
                    vuetify.VTextField(
                        classes="mb-1",
                        dense=True,
                        hide_details=True,
                        readonly=True,
                        outlined=True,
                        disabled=True,
                        type="string",
                        v_model=("directory_label", None),
                    )


class FileSelector(AbstractCard):
    def __init__(self, file_selection):
        super().__init__("file", "mdi-application-outline", "Browse")
        with self._content as content:
            content.classes = "px-0 mx-0"
            content.style = "max-height: min(12rem, 40vh); overflow: auto;"
            with vuetify.VList(
                dense=True,
                hide_details=True,
            ):
                with vuetify.VListItemGroup(
                    v_model=("file_selected", None),
                    color="primary",
                ):
                    with vuetify.VListItem(
                        v_for="(item, index) in files",
                        key="index",
                        dense=True,
                        hide_details=True,
                        click=(file_selection, "[item]"),
                        color="primary",
                        style="min-height:10px",
                    ):
                        with vuetify.VListItemIcon(classes="ma-0 mr-1"):
                            vuetify.VIcon("mdi-file", x_small=True)
                        with vuetify.VListItemContent(classes="pa-0"):
                            vuetify.VListItemTitle("{{item}}")


class FigureControl(AbstractCard):
    def __init__(self):
        super().__init__("viz", "mdi-image-multiple-outline", "Image")
        with self._content:
            with vuetify.VRow(classes="pa-0 pt-1", dense=True, hide_details=True):
                with vuetify.VCol(cols="6"):
                    vuetify.VSelect(
                        v_model=("selectedRepresentation", "Heatmap"),
                        items=("representations", ["Heatmap", "Contours", "Combined"]),
                        label="Representation",
                        dense=True,
                        hide_details=True,
                        outlined=True,
                    )
                with vuetify.VCol(cols="6"):
                    vuetify.VSelect(
                        v_model=("selectedColor", "spectral"),
                        items=(
                            "colors",
                            ["spectral", "rdbu", "gray", "ylgn", "ylorbr"],
                        ),
                        label="Colors",
                        dense=True,
                        hide_details=True,
                        outlined=True,
                    )
            with vuetify.VRow(classes="pa-0 pt-1", dense=True, hide_details=True):
                with vuetify.VCol(cols="6"):
                    vuetify.VCheckbox(
                        v_model=("contour_labels", True),
                        label="Contour labels",
                        dense=True,
                        hide_details=True,
                    )
                with vuetify.VCol(cols="6"):
                    vuetify.VCheckbox(
                        v_model=("show_mask", True),
                        label="Show mask",
                        dense=True,
                        hide_details=True,
                    )


class Scattering(AbstractCard):
    def __init__(self):
        super().__init__("scattering", "mdi-scatter-plot-outline", "Scattering")
        with self._content:
            create_scattering_content()


class Transmission(AbstractCard):
    def __init__(self):
        super().__init__("transmission", "mdi-transfer", "Transmission")
        with self._content:
            create_transmission_content()



import os


def get_directory_structure(rootdir):
    """
    Creates a nested dictionary that represents the folder structure of rootdir
    """
    dirlist = []
    rootdir = rootdir.rstrip(os.sep)
    for path, dirs, _ in os.walk(rootdir):
        sublist = []
        for dir in dirs:
            item = {
                "name": dir,
                "type": "dir",
                "path": path + "/" + dir,
                "children": None,
            }
            sublist.append(item)
        if not search_replace(dirlist, path, sublist):
            for item in sublist:
                dirlist.append(item)
    return dirlist


def get_file_list(path):
    """
    Creates a list of files in a folder
    """
    filelist = os.listdir(path)
    return filelist


def search_replace(data, path, subdata):
    found = False
    for item in data:
        if item["path"] == path:
            item["children"] = subdata
            found = True
            break
    return found



from trame.decorators import TrameApp

import plotly.graph_objects as go
from plotly.subplots import make_subplots


@TrameApp()
class PlotlyDetectorView:
    def __init__(self, server):
        self.server = server

        self.color_range = [0, 1]
        self._left_data = None
        self._left_mask = None
        self._left_x = None
        self._left_y = None
        self._center_data = None
        self._center_mask = None
        self._center_x = None
        self._center_y = None
        self._right_data = None
        self._right_mask = None
        self._right_x = None
        self._right_y = None

    def set_data_minmax(self, data_min, data_max):
        self.color_range = [data_min, data_max]

    def update_left_data(self, data, mask, x, y, **kwargs):
        self._left_data = data
        self._left_mask = mask
        self._left_x = x
        self._left_y = y

    def update_center_data(self, data, mask, x, y, **kwargs):
        self._center_data = data
        self._center_mask = mask
        self._center_x = x
        self._center_y = y

    def update_right_data(self, data, mask, x, y, **kwargs):
        self._right_data = data
        self._right_mask = mask
        self._right_x = x
        self._right_y = y

    def update_d11_view(self, **kwargs):
        state = self.server.state

        fig = make_subplots(
            rows=5,
            cols=4,
            specs=[
                [None, {"rowspan": 5, "colspan": 2}, None, None],
                [
                    {"rowspan": 3, "colspan": 1},
                    None,
                    None,
                    {"rowspan": 3, "colspan": 1},
                ],
                [None, None, None, None],
                [None, None, None, None],
                [None, None, None, None],
            ],
            print_grid=False,
        )
        # center (1,2), left (2,1), right (2,4)

        # center
        if state.selectedRepresentation == "Heatmap":
            self.create_heatmap(fig, self._center_data, self._center_mask, 1, 2)
        elif state.selectedRepresentation == "Contours":
            self.create_contour(fig, self._center_data, 1, 2)
        else:
            self.create_combined(fig, self._center_data, self._center_mask, 1, 2)
        fig.update_yaxes(
            range=(0.0, state.center_nx),
            constrain="domain",
            scaleanchor="x",
            scaleratio=state.pixel_ratio,
            row=1,
            col=2,
        )
        fig.update_xaxes(
            range=(0.0, state.center_ny), constrain="domain", scaleratio=2, row=1, col=2
        )

        # left
        if state.selectedRepresentation == "Heatmap":
            self.create_heatmap(fig, self._left_data, self._left_mask, 2, 1)
        elif state.selectedRepresentation == "Contours":
            self.create_contour(fig, self._left_data, 2, 1)
        else:
            self.create_combined(fig, self._left_data, self._left_mask, 2, 1)
        fig.update_yaxes(range=(0.0, state.left_nx), constrain="domain", row=2, col=1)
        fig.update_xaxes(
            range=(0.0, state.left_ny),
            constrain="domain",
            scaleanchor="y",
            scaleratio=1,
            row=2,
            col=1,
        )

        # right
        if state.selectedRepresentation == "Heatmap":
            self.create_heatmap(fig, self._right_data, self._right_mask, 2, 4)
        elif state.selectedRepresentation == "Contours":
            self.create_contour(fig, self._right_data, 2, 4)
        else:
            self.create_combined(fig, self._right_data, self._right_mask, 2, 4)
        fig.update_yaxes(range=(0.0, state.right_nx), constrain="domain", row=2, col=4)
        fig.update_xaxes(
            range=(0.0, state.right_ny),
            constrain="domain",
            scaleanchor="y",
            scaleratio=1,
            row=2,
            col=4,
        )
        fig.update_layout(
            coloraxis=dict(colorscale=state.selectedColor), showlegend=False
        )
        fig.update_coloraxes(
            colorbar={
                "orientation": "h",
                "lenmode": "fraction",
                "len": 0.6,
                "thickness": 30,
                "y": -0.25,
            }
        )

        self.server.controller.plotly_view_update(fig)

        return fig

    def update_cg2_view(self, **kwargs):
        state = self.server.state

        fig = make_subplots(
            rows=1,
            cols=1,
            specs=[[{"rowspan": 1, "colspan": 1}]],
            print_grid=False,
        )
        # center (1,1)

        # center
        if state.selectedRepresentation == "Heatmap":
            self.create_heatmap(fig, self._center_data, self._center_mask, 1, 1)
        elif state.selectedRepresentation == "Contours":
            self.create_contour(fig, self._center_data, 1, 1)
        else:
            self.create_combined(fig, self._center_data, self._center_mask, 1, 1)
        fig.update_yaxes(
            range=(0.0, state.center_nx),
            constrain="domain",
            scaleanchor="x",
            scaleratio=state.pixel_ratio,
            row=1,
            col=1,
        )
        fig.update_xaxes(
            range=(0.0, state.center_ny), constrain="domain", scaleratio=2, row=1, col=1
        )
        fig.update_layout(
            coloraxis=dict(colorscale=state.selectedColor), showlegend=False
        )
        fig.update_coloraxes(
            colorbar={
                "orientation": "h",
                "lenmode": "fraction",
                "len": 0.6,
                "thickness": 30,
                "y": -0.25,
            }
        )

        self.server.controller.plotly_view_update(fig)

        return fig

    def create_heatmap(self, fig, data, mask, _row, _col, **kwargs):
        state = self.server.state
        fig.add_trace(
            go.Heatmap(
                z=data,
                opacity=1.0,
                coloraxis="coloraxis",
            ),
            row=_row,
            col=_col,
        )
        if state.show_mask:
            fig.append_trace(
                go.Heatmap(
                    z=mask,
                    opacity=1.0,
                    colorscale=[[0, "rgb(0,0,0)"], [1, "rgb(0,0,0)"]],
                    showlegend=False,
                    showscale=False,
                ),
                row=_row,
                col=_col,
            )

    def create_contour(self, fig, data, _row, _col, **kwargs):
        state = self.server.state

        fig.add_trace(
            go.Contour(
                z=data,
                contours={
                    "coloring": "lines",
                    "showlabels": state.contour_labels,
                    "labelfont": {
                        "size": 12,
                        "color": "black",
                    },
                },
                contours_coloring="lines",
                line_smoothing=1,
                coloraxis="coloraxis",
                name="(1,2)",
            ),
            row=_row,
            col=_col,
        )

    def create_combined(self, fig, data, mask, _row, _col, **kwargs):
        state = self.server.state

        fig.add_trace(
            go.Contour(
                z=data,
                contours={
                    "coloring": "heatmap",
                    "showlabels": state.contour_labels,
                    "labelfont": {
                        "size": 12,
                        "color": "black",
                    },
                },
                line_smoothing=1,
                coloraxis="coloraxis",
                name="(1,2)",
            ),
            row=_row,
            col=_col,
        )
        if state.show_mask:
            fig.append_trace(
                go.Heatmap(
                    z=mask,
                    opacity=1.0,
                    colorscale=[[0, "rgb(0,0,0)"], [1, "rgb(0,0,0)"]],
                    showlegend=False,
                    showscale=False,
                ),
                row=_row,
                col=_col,
            )



import os
import logging

from trame.app import get_server
from trame.decorators import TrameApp, change
from trame.ui.vuetify import SinglePageWithDrawerLayout
from trame.widgets import vuetify, plotly, vtk

from .assets import ASSETS
from .ui import (
    DeviceSelector,
    Directory,
    FileSelector,
    FigureControl,
    Scattering,
    Transmission,
)
from .instrument.d11_plus import D11_Plus
from .instrument.cg2 import CG2
from .visualization import Visualization
from .utilities import file_search as fs

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

PANELS = [
    ("device", "mdi-microscope"),
    ("directory", "mdi-folder-eye"),
    ("file", "mdi-application"),
    ("viz", "mdi-image-multiple"),
    ("scattering", "mdi-scatter-plot"),
    ("transmission", "mdi-transfer"),
]


# ---------------------------------------------------------
# AirSans class
# ---------------------------------------------------------


@TrameApp()
class AirSans:
    def __init__(self, server=None, data=None):
        self.server = get_server(server, client_type="vue2")
        self._viz = Visualization(self.server)
        self._selected_device = None
        self._active_directory = None

        # CLI
        self.server.cli.add_argument(
            "--data",
            help="Path to browse",
            dest="data",
            default=".",
        )
        args, _ = self.server.cli.parse_known_args()
        if data is None:
            data = args.data

        # Detectors
        self.state.left_detector = False
        self.state.center_detector = False
        self.state.right_detector = False

        # search directory contents
        self.state.dirs = fs.get_directory_structure(data)
        self.state.files = []
        self.state.file = None

        # Initialize data
        self.state.data = None
        self.state.center_data = None
        self.state.pixel_ratio = 2.0

        self.ui = self._build_ui()

    @property
    def state(self):
        return self.server.state

    @property
    def ctrl(self):
        return self.server.controller

    def _build_ui(self):
        self.state.trame__title = "air-sans"
        with SinglePageWithDrawerLayout(self.server) as layout:
            with layout.icon:
                vuetify.VIcon("mdi-bullseye")

            with layout.title as title:
                title.set_text("AIR-SANS")
                title.classes = "font-weight-medium ml-0 pl-0"

            with layout.toolbar as tb:
                tb.dense = True
                with vuetify.VBtnToggle(
                    v_model=("panel_visible", ["device", "directory", "file", "viz"]),
                    multiple=True,
                    dense=True,
                    classes="ml-6",
                ):
                    for key, icon in PANELS:
                        with vuetify.VBtn(value=key, small=True):
                            vuetify.VIcon(icon, small=True)

                vuetify.VSpacer()
                with vuetify.VBtnToggle(
                    dense=True,
                    classes="ml-6",
                    v_model=("renderer", "plotly"),
                    hide_details=True,
                    mandatory=True,
                ):
                    with vuetify.VBtn(
                        small=True,
                        icon=True,
                        value="plotly",
                    ):
                        vuetify.VImg(
                            src=ASSETS.plotly, height=20, width=20, contain=True
                        )
                    with vuetify.VBtn(
                        small=True,
                        icon=True,
                        value="vtk",
                    ):
                        vuetify.VImg(src=ASSETS.vtk, height=20, width=20, contain=True)

            with layout.drawer as drawer:
                drawer.width = 350

                DeviceSelector()
                Directory(select_directory_fn=self.select_directory)
                FileSelector(self.selected_file)
                FigureControl()
                Scattering()
                Transmission()

            with layout.content:
                with vuetify.VContainer(
                    fluid=True,
                    classes="pa-0 fill-height",
                ):
                    with vuetify.VRow(
                        v_show="selectedDevice == 'CG2' && renderer == 'vtk'",
                        classes="fill-height pa-0",
                    ):
                        with vuetify.VCol(
                            cols=12, v_show="center_detector", classes="pa-0"
                        ):
                            view = vtk.VtkRemoteView(
                                self._viz.full_render_window, interactive_ratio=1
                            )
                            self._viz.set_full_view(view)
                    with vuetify.VRow(
                        v_show="selectedDevice == 'D11+' && renderer == 'vtk'",
                        classes="fill-height pa-0",
                    ):
                        with vuetify.VCol(
                            cols=3, v_show="left_detector", classes="pa-0"
                        ):
                            view = vtk.VtkRemoteView(
                                self._viz.left_render_window, interactive_ratio=1
                            )
                            self._viz.set_left_view(view)
                        with vuetify.VCol(
                            cols=6, v_show="center_detector", classes="pa-0"
                        ):
                            view = vtk.VtkRemoteView(
                                self._viz.center_render_window, interactive_ratio=1
                            )
                            self._viz.set_center_view(view)
                        with vuetify.VCol(
                            cols=3,
                            v_show="right_detector",
                            classes="pa-0",
                        ):
                            view = vtk.VtkRemoteView(
                                self._viz.right_render_window, interactive_ratio=1
                            )
                            self._viz.set_right_view(view)
                    with vuetify.VRow(
                        v_show="renderer == 'plotly'", classes="fill-height pa-0"
                    ):
                        self.ctrl.plotly_view_update = plotly.Figure(
                            display_mode_bar=("false",),
                            v_show=("figure_ready", False),
                        ).update

            return layout

    @change("selectedDevice")
    def on_device_change(self, selectedDevice, **kwargs):
        if selectedDevice == "D11+":
            self._viz.set_selected_device("D11+")
            self._selected_device = D11_Plus()
            self.state.left_detector = True
            self.state.center_detector = True
            self.state.right_detector = True
            self._viz.set_center_mask(self._selected_device.detector1_mask)
            self._viz.set_left_mask(self._selected_device.detector2_mask)
            self._viz.set_right_mask(self._selected_device.detector3_mask)
        elif selectedDevice == "CG2":
            self._viz.set_selected_device("CG2")
            self._selected_device = CG2()
            self.state.left_detector = True
            self.state.center_detector = True
            self.state.right_detector = True
            self._viz.set_center_mask(self._selected_device.detector1_mask)

    def select_directory(self, active_nodes):
        self.server.state.directory_label = None
        self._active_directory = None
        if len(active_nodes):
            node_id = active_nodes[0]
            if isinstance(node_id, str):
                self.server.state.directory_label = os.path.basename(node_id)
                self._active_directory = node_id
                self.server.state.files = fs.get_file_list(node_id)

    def selected_file(self, file):
        if self._selected_device.Instrument == "D11+":
            self.open_d11plus_file(file)
        elif self._selected_device.Instrument == "CG2":
            self.open_cg2_file(file)
        else:
            return

    def open_cg2_file(self, file):
        state = self.server.state
        state.file = file
        self._selected_device.read_file(self._active_directory, state.file)
        self._viz.reset_data_minmax()
        pixel1_x = self._selected_device.detector1_pixel_size[0]
        pixel1_y = self._selected_device.detector1_pixel_size[1]
        state.pixel_ratio = pixel1_y / pixel1_x
        state.center_ny = self._selected_device.ny1
        state.center_nx = self._selected_device.nx1
        self._viz.set_center_data(self._selected_device.det1_data, pixel1_x, pixel1_y)

        self._viz.update_visualization()

        state.figure_ready = True

    def open_d11plus_file(self, file):
        state = self.server.state
        state.file = file
        self._selected_device.read_file(self._active_directory, state.file)
        self._viz.reset_data_minmax()

        pixel1_x = self._selected_device.pixel1_x
        pixel1_y = self._selected_device.pixel1_y
        state.pixel_ratio = pixel1_y / pixel1_x
        state.center_ny = self._selected_device.ny1
        state.center_nx = self._selected_device.nx1
        self._viz.set_center_data(self._selected_device.det1_data, pixel1_x, pixel1_y)

        pixel2_x = self._selected_device.pixel2_x
        pixel2_y = self._selected_device.pixel2_y
        state.left_ny = self._selected_device.ny2
        state.left_nx = self._selected_device.nx2
        self._viz.set_left_data(self._selected_device.det2_data, pixel2_x, pixel2_y)

        pixel3_x = self._selected_device.pixel3_x
        pixel3_y = self._selected_device.pixel3_y
        state.right_ny = self._selected_device.ny3
        state.right_nx = self._selected_device.nx3
        self._viz.set_right_data(self._selected_device.det3_data, pixel3_x, pixel3_y)

        self._viz.update_visualization()

        state.figure_ready = True

    @change("device_active_data")
    def show_imask(self, device_active_data, **kwargs):
        self.server.state.figure_ready = False
        if device_active_data == "mask":
            pixel1_x = self._selected_device.pixel1_x
            pixel1_y = self._selected_device.pixel1_y
            pixel2_x = self._selected_device.pixel2_x
            pixel2_y = self._selected_device.pixel2_y
            pixel3_x = self._selected_device.pixel3_x
            pixel3_y = self._selected_device.pixel3_y
            self._viz.set_center_data(
                self._selected_device.detector1_imask_data, pixel1_x, pixel1_y
            )
            self._viz.set_left_data(
                self._selected_device.detector2_imask_data, pixel2_x, pixel2_y
            )
            self._viz.set_right_data(
                self._selected_device.detector3_imask_data, pixel3_x, pixel3_y
            )
            self.server.state.figure_ready = True
            self._viz.update_visualization()

        if device_active_data == "efficiency":
            pixel1_x = self._selected_device.pixel1_x
            pixel1_y = self._selected_device.pixel1_y
            pixel2_x = self._selected_device.pixel2_x
            pixel2_y = self._selected_device.pixel2_y
            pixel3_x = self._selected_device.pixel3_x
            pixel3_y = self._selected_device.pixel3_y
            self._viz.set_center_data(
                self._selected_device.detector1_efficiency_data, pixel1_x, pixel1_y
            )
            self._viz.set_left_data(
                self._selected_device.detector2_efficiency_data, pixel2_x, pixel2_y
            )
            self._viz.set_right_data(
                self._selected_device.detector3_efficiency_data, pixel3_x, pixel3_y
            )
            self.server.state.figure_ready = True
            self._viz.update_visualization()

        if device_active_data == "error":
            pixel1_x = self._selected_device.pixel1_x
            pixel1_y = self._selected_device.pixel1_y
            pixel2_x = self._selected_device.pixel2_x
            pixel2_y = self._selected_device.pixel2_y
            pixel3_x = self._selected_device.pixel3_x
            pixel3_y = self._selected_device.pixel3_y
            self._viz.set_center_data(
                self._selected_device.detector1_efficiency_error, pixel1_x, pixel1_y
            )
            self._viz.set_left_data(
                self._selected_device.detector2_efficiency_error, pixel2_x, pixel2_y
            )
            self._viz.set_right_data(
                self._selected_device.detector3_efficiency_error, pixel3_x, pixel3_y
            )
            self.server.state.figure_ready = True
            self._viz.update_visualization()

        if device_active_data == "" and self.server.state.file:
            self.selected_file(self.server.state.file)



import numpy as np

from trame.decorators import TrameApp, change

from .vtk import VTKDetectorView
from .plotly import PlotlyDetectorView


@TrameApp()
class Visualization:
    def __init__(self, server):
        self.server = server
        self.selected_device = None
        self._data_max = -2147483648
        self._data_min = 2147483648
        self._left_data = None
        self._left_mask = None
        self._left_x = None
        self._left_y = None
        self._center_data = None
        self._center_mask = None
        self._center_x = None
        self._center_y = None
        self._right_data = None
        self._right_mask = None
        self._right_x = None
        self._right_y = None
        # VTK Views
        self._full_view = VTKDetectorView(server, True)
        self._left_view = VTKDetectorView(server, False)
        self._center_view = VTKDetectorView(server, True)
        self._right_view = VTKDetectorView(server, False)
        # Plotly Views
        self._all_view = PlotlyDetectorView(server)

    def set_selected_device(self, device):
        self.selected_device = device

    def reset_data_minmax(self):
        self._data_max = -2147483648
        self._data_min = 2147483648

    def set_left_data(self, data, x, y):
        self._left_data = data
        tmp_min = np.min(data)
        tmp_max = np.max(data)
        self._data_min = min(self._data_min, tmp_min)
        self._data_max = max(self._data_max, tmp_max)
        self._left_x = x
        self._left_y = y

    def set_center_data(self, data, x, y):
        self._center_data = data
        tmp_min = np.min(data)
        tmp_max = np.max(data)
        self._data_min = min(self._data_min, tmp_min)
        self._data_max = max(self._data_max, tmp_max)
        self._center_x = x
        self._center_y = y

    def set_right_data(self, data, x, y):
        self._right_data = data
        tmp_min = np.min(data)
        tmp_max = np.max(data)
        self._data_min = min(self._data_min, tmp_min)
        self._data_max = max(self._data_max, tmp_max)
        self._right_x = x
        self._right_y = y

    def set_left_mask(self, mask):
        self._left_mask = mask

    def set_center_mask(self, mask):
        self._center_mask = mask

    def set_right_mask(self, mask):
        self._right_mask = mask

    @change(
        "renderer",
        "selectedRepresentation",
        "contour_labels",
        "show_mask",
        "selectedColor",
        "center_nx",
        "center_ny",
        "pixel_ratio",
        "left_nx",
        "left_ny",
        "right_nx",
        "right_ny",
    )
    def update_visualization(self, **kwargs):
        if self.selected_device == "D11+":
            self.update_d11_visualization()
        elif self.selected_device == "CG2":
            self.update_cg2_visualization()

    def update_d11_visualization(self, **kwargs):
        if self.server.state.renderer != "plotly":
            self.update_d11_vtk()
        else:
            self.update_d11_plotly()

    def update_cg2_visualization(self, **kwargs):
        if self.server.state.renderer != "plotly":
            self.update_cg2_vtk()
        else:
            self.update_cg2_plotly()

    ################################
    # VTK
    ################################

    @property
    def full_render_window(self):
        return self._full_view.render_window

    def set_full_view(self, html_view):
        self._full_view.set_html_view(html_view)

    @property
    def left_render_window(self):
        return self._left_view.render_window

    def set_left_view(self, html_view):
        self._left_view.set_html_view(html_view)

    @property
    def center_render_window(self):
        return self._center_view.render_window

    def set_center_view(self, html_view):
        self._center_view.set_html_view(html_view)

    @property
    def right_render_window(self):
        return self._right_view.render_window

    def set_right_view(self, html_view):
        self._right_view.set_html_view(html_view)

    def reset_camera(self):
        self._left_view.html_reset_camera()
        self._center_view.html_reset_camera()
        self._right_view.html_reset_camera()

    def update_d11_vtk(self, **kwargs):
        self._left_view.set_data_minmax(self._data_min, self._data_max)
        self._left_view.update_data(
            self._left_data, self._left_mask, self._left_x, self._left_y
        )
        self._center_view.set_data_minmax(self._data_min, self._data_max)
        self._center_view.update_data(
            self._center_data, self._center_mask, self._center_x, self._center_y
        )
        self._right_view.set_data_minmax(self._data_min, self._data_max)
        self._right_view.update_data(
            self._right_data, self._right_mask, self._right_x, self._right_y
        )

    def update_cg2_vtk(self, **kwargs):
        self._full_view.set_data_minmax(self._data_min, self._data_max)
        self._full_view.update_data(
            self._center_data, self._center_mask, self._center_x, self._center_y
        )

    ################################
    # Plotly
    ################################

    def update_d11_plotly(self, **kwargs):
        self._all_view.set_data_minmax(self._data_min, self._data_max)
        self._all_view.update_left_data(
            self._left_data, self._left_mask, self._left_x, self._left_y
        )
        self._all_view.update_center_data(
            self._center_data, self._center_mask, self._center_x, self._center_y
        )
        self._all_view.update_right_data(
            self._right_data, self._right_mask, self._right_x, self._right_y
        )
        self._all_view.update_d11_view()

    def update_cg2_plotly(self, **kwargs):
        self._all_view.set_data_minmax(self._data_min, self._data_max)
        self._all_view.update_center_data(
            self._center_data, self._center_mask, self._center_x, self._center_y
        )
        self._all_view.update_cg2_view()



import numpy as np
import h5py

from pathlib import Path
from . import mat_files as mf

BASE = Path(__file__).parent


class D11_Plus:
    def __init__(self):
        self.Facility = "ILL"
        self.Instrument = "D11+"
        self.Name = "D11+_(2021-current)"
        self.Highlight = True
        # Filename
        self.filename_numeric_length = 6
        self.filename_lead_string = []
        self.filename_tail_string = []
        self.filename_extension_string = [".nxs"]
        self.filename_data_loader = "raw_read_ill_nexus_d11d22_2021"
        # Misc
        self.guide_size = [45e-3, 45e-3]  # [x,y]
        self.attenuators = [1, 112.8, 421.14, 1911.96]  # Attenautors 0, 1, 2, 3 etc.
        # Detector(s)
        self.detectors = 3
        # Main Panel
        self.detector1_type = "tube"
        self.detector1_view_position = "centre"
        self.detector1_name = "Main Panel"
        self.detector1_det_offset = (
            0  # position difference between panel and main detector
        )
        self.detector1_pixels = [256, 192]  # Number of pixels, x,y
        self.detector1_pixel_size = [4, 8]
        # mm [x y]
        self.detector1_nominal_beam_centre = [128.5, 103.5]  # [x y]
        self.detector1_nominal_det_translation = [0, 0]  # mm [x y]
        self.detector1_dead_time = 1.5e-6 * np.ones(
            (192, 1)
        )  # the direction of this matrix determines the direction of the tubes
        self.detector1_imask_data = mf.read_mask_mat_file(
            BASE / "det_mask" / "det1_d11_plus.mat"
        )
        self.reset_detector1_mask()
        self.detector1_relative_efficiency = (
            1  # Efficiency relative to rear detector (detector 1)
        )
        self.detector1_efficiency_data = mf.read_efficiency_mat_file_data(
            BASE / "det_efficiency" / "det1_d11_plus.mat"
        )
        self.detector1_efficiency_error = mf.read_efficiency_mat_file_error(
            BASE / "det_efficiency" / "det1_d11_plus.mat"
        )
        # Left Panel
        self.detector2_type = "tube"
        self.detector2_view_position = "left"
        self.detector2_name = "Left"
        self.detector2_det_offset = (
            -0.105
        )  # position difference between panel and main detector
        self.detector2_pixels = [32, 256]  # Number of pixels, x,y
        self.detector2_pixel_size = [8, 4]  # mm [x y]
        self.detector2_nominal_beam_centre = [16.5, 142.5]  # [x y]
        self.detector2_nominal_det_translation = [-628, 0]  # mm [x y]
        self.detector2_dead_time = 1.5e-6 * np.ones((1, 32))
        self.detector2_imask_data = mf.read_mask_mat_file(
            BASE / "det_mask" / "det2_d11_plus.mat"
        )
        self.reset_detector2_mask()
        self.detector2_relative_efficiency = 1
        self.detector2_efficiency_data = mf.read_efficiency_mat_file_data(
            BASE / "det_efficiency" / "det2_d11_plus.mat"
        )
        self.detector2_efficiency_error = mf.read_efficiency_mat_file_error(
            BASE / "det_efficiency" / "det2_d11_plus.mat"
        )
        # Right Panel
        self.detector3_type = "tube"
        self.detector3_view_position = "right"
        self.detector3_name = "Right"
        self.detector3_det_offset = (
            -0.105
        )  # position difference between panel and main detector
        self.detector3_pixels = [32, 256]  # Number of pixels, x,y
        self.detector3_pixel_size = [8, 4]  # mm [x y]
        self.detector3_nominal_beam_centre = [16.5, 142.5]  # [x y]
        self.detector3_nominal_det_translation = [+628, 0]  # mm [x y]
        self.detector3_dead_time = 1.5e-6 * np.ones((1, 32))
        self.detector3_imask_data = mf.read_mask_mat_file(
            BASE / "det_mask" / "det3_d11_plus.mat"
        )
        self.reset_detector3_mask()
        self.detector3_relative_efficiency = 1
        self.detector3_efficiency_data = mf.read_efficiency_mat_file_data(
            BASE / "det_efficiency" / "det3_d11_plus.mat"
        )
        self.detector3_efficiency_error = mf.read_efficiency_mat_file_error(
            BASE / "det_efficiency" / "det3_d11_plus.mat"
        )

    def reset_detector1_mask(self):
        self.detector1_mask = np.where(
            self.detector1_imask_data == 1, np.nan, self.detector1_imask_data
        )

    def reset_detector2_mask(self):
        self.detector2_mask = np.where(
            self.detector2_imask_data == 1, np.nan, self.detector2_imask_data
        )

    def reset_detector3_mask(self):
        self.detector3_mask = np.where(
            self.detector3_imask_data == 1, np.nan, self.detector3_imask_data
        )

    def read_file(self, dir, path):
        self.path = dir + "/" + path
        root = h5py.File(self.path, "r")
        keys = list(root.keys())
        entry = root.name + keys[0]
        self.read_parameters(root, entry)

    def read_parameters(self, root, entry):
        # Run
        self.run_number = self.read_value(root, entry, "/run_number")
        self.user = self.read_value(root, entry, "/user/name")
        self.start_time = self.read_value(root, entry, "/start_time")
        self.end_time = self.read_value(root, entry, "/end_time")
        self.duration = self.read_value(root, entry, "/duration")
        self.experiment_identifier = self.read_value(
            root, entry, "/experiment_identifier"
        )
        self.experiment_title = self.read_value(root, entry, "/experiment_title")
        self.sample_description = self.read_value(root, entry, "/sample_description")
        # Detector Distance & Motors
        self.pixel1_y = self.read_value(root, entry, "/D11/Detector 1/pixel_size_y")
        self.pixel1_x = self.read_value(root, entry, "/D11/Detector 1/pixel_size_x")
        # Detector rate
        self.det1_rate = self.read_value(root, entry, "/D11/Detector 1/det1_rate")
        # Detector motors
        self.det1_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 1/det1_actual",
            fallback_item="/D11/Detector 1/det_actual",
        )
        self.det1_calc = self.read_value(
            root,
            entry,
            "/D11/Detector 1/det1_calc",
            fallback_item="/D11/Detector 1/det_calc",
        )
        self.dan1_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 1/dan1_actual",
            fallback_item="/D11/Detector 1/dan_actual",
        )
        self.dtr1_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 1/dtr1_actual",
            fallback_item="/D11/Detector 1/dtr_actual",
        )
        # BeamStop
        self.bx1_actual = self.read_value(
            root,
            entry,
            "/D11//beamstop/bx1_actual",
            fallback_item="/D11//beamstop/bx_actual",
            default=0.0,
        )
        self.by1_actual = self.read_value(
            root,
            entry,
            "/D11//beamstop/by1_actual",
            fallback_item="/D11//beamstop/by_actual",
            default=0.0,
        )

        self.pixel2_y = self.read_value(root, entry, "/D11/Detector 2/pixel_size_y")
        self.pixel2_x = self.read_value(root, entry, "/D11/Detector 2/pixel_size_x")
        # Detector rate
        self.det2_rate = self.read_value(root, entry, "/D11/Detector 2/det2_rate")
        # Detector motors
        self.det2_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 2/det2_actual",
            fallback_item="/D11/Detector 2/det_actual",
        )
        self.det2_calc = self.read_value(
            root,
            entry,
            "/D11/Detector 2/det2_calc",
            fallback_item="/D11/Detector 2/det_calc",
        )
        self.dan2_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 2/dan2_actual",
            fallback_item="/D11/Detector 2/dan_actual",
        )
        self.dtr2_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 2/dtr2_actual",
            fallback_item="/D11/Detector 2/dtr_actual",
        )
        # BeamStop
        self.bx2_actual = self.read_value(
            root,
            entry,
            "/D11//beamstop/bx2_actual",
            fallback_item="/D11//beamstop/bx_actual",
            default=0.0,
        )
        self.by2_actual = self.read_value(
            root,
            entry,
            "/D11//beamstop/by2_actual",
            fallback_item="/D11//beamstop/by_actual",
            default=0.0,
        )

        self.pixel3_y = self.read_value(root, entry, "/D11/Detector 3/pixel_size_y")
        self.pixel3_x = self.read_value(root, entry, "/D11/Detector 3/pixel_size_x")
        # Detector rate
        self.det3_rate = self.read_value(root, entry, "/D11/Detector 3/det3_rate")
        # Detector motors
        self.det3_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 3/det3_actual",
            fallback_item="/D11/Detector 3/det_actual",
        )
        self.det3_calc = self.read_value(
            root,
            entry,
            "/D11/Detector 3/det3_calc",
            fallback_item="/D11/Detector 3/det_calc",
        )
        self.dan3_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 3/dan3_actual",
            fallback_item="/D11/Detector 3/dan_actual",
        )
        self.dtr3_actual = self.read_value(
            root,
            entry,
            "/D11/Detector 3/dtr3_actual",
            fallback_item="/D11/Detector 3/dtr_actual",
        )
        # BeamStop
        self.bx3_actual = self.read_value(
            root,
            entry,
            "/D11//beamstop/bx3_actual",
            fallback_item="/D11//beamstop/bx_actual",
            default=0.0,
        )
        self.by3_actual = self.read_value(
            root,
            entry,
            "/D11//beamstop/by3_actual",
            fallback_item="/D11//beamstop/by_actual",
            default=0.0,
        )
        # Wavelength
        self.wavelength = self.read_value(root, entry, "/D11/selector/wavelength")
        self.wavelength_res = (
            self.read_value(root, entry, "/D11/selector/wavelength_res", default=10)
            / 100
        )
        self.rotation_speed = self.read_value(
            root, entry, item="/D11/selector/rotation_speed"
        )
        self.seltrs_actual = self.read_value(root, entry, "/D11/selector/seltrs_actual")

        # Attenuator
        self.attenuator_position = self.read_value(
            root, entry, "/D11/attenuator/position", default=0
        )
        self.attenuation_coefficient = self.read_value(
            root, entry, "/D11/attenuator/attenuation_coefficient"
        )
        self.attenuation_value = self.read_value(
            root,
            entry,
            "/D11/attenuator/attenuation_value",
            default=[
                self.attenuators[1],
                self.attenuators[2],
                self.attenuators[3],
            ],
        )
        # Collimation
        self.collimation_actual_position = self.read_value(
            root, entry, "/D11/collimation/actual_position"
        )
        # Source Size
        temp = self.read_array(
            root,
            entry,
            "/D11/collimation/ap_size",
            default=[self.guide_size[0] * 1000, self.guide_size[1] * 1000],
        )
        self.collimation_ap_size_x = temp[0] / 1000
        self.collimation_ap_size_y = temp[1] / 1000
        # Reactor Power
        self.reactor_power = self.read_value(root, entry, "/reactor_power")
        # Sample
        self.sample_thickness = self.read_value(root, entry, "/sample/thickness")
        # Sample Motors
        self.sample_san_actual = self.read_value(root, entry, "/sample/san_actual")
        self.sample_phi_actual = self.read_value(root, entry, "/sample/phi_actual")
        self.sample_sdi1_actual = self.read_value(root, entry, "/sample/sdi1_actual")
        self.sample_sdi_actual = self.read_value(root, entry, "/sample/sdi_actual")
        self.sample_sdi2_actual = self.read_value(root, entry, "/sample/sdi2_actual")
        self.sample_trs_actual = self.read_value(root, entry, "/sample/trs_actual")
        self.sample_sht_actual = self.read_value(root, entry, "/sample/sht_actual")
        self.sample_str_actual = self.read_value(root, entry, "/sample/str_actual")
        self.sample_changer_value = self.read_value(
            root, entry, "/sample/sample_changer_value"
        )
        self.sample_omega_actual = self.read_value(root, entry, "/sample/omega_actual")
        self.sample_TrMicro_actual = self.read_value(
            root, entry, "/sample/TrMicro_actual"
        )
        self.sample_TrMicro_offset = self.read_value(
            root, entry, "/sample/TrMicro_offset"
        )
        self.sample_TrMicro_requested = self.read_value(
            root, entry, "/sample/TrMicro_requested"
        )

        # Sample environment
        self.sample_temperature = self.read_value(root, entry, "/sample/temperature")
        self.sample_regulation_temperature = self.read_value(
            root, entry, "/sample/regulation_temperature"
        )
        self.sample_setpoint_temperature = self.read_value(
            root, entry, "/sample/setpoint_temperature"
        )
        self.sample_bath1_regulation_temperature = self.read_value(
            root, entry, "/sample/bath1_regulation_temperature"
        )
        self.sample_bath1_setpoint_temperature = self.read_value(
            root, entry, "/sample/bath1_setpoint_temperature"
        )
        self.sample_bath2_regulation_temperature = self.read_value(
            root, entry, "/sample/bath2_regulation_temperature"
        )
        self.sample_bath2_setpoint_temperature = self.read_value(
            root, entry, "/sample/bath2_setpoint_temperature"
        )
        self.sample_bath_selector_actual = self.read_value(
            root, entry, "/sample/bath_selector_actual"
        )
        self.sample_air_temperature = self.read_value(
            root, entry, "/sample/air_temperature"
        )
        self.sample_rack_temperature = self.read_value(
            root, entry, "/sample/rack_temperature"
        )

        # Power Supplies
        self.sample_ps1_current = self.read_value(root, entry, "/sample/ps1_current")
        self.sample_ps1_voltage = self.read_value(root, entry, "/sample/ps1_voltage")
        self.sample_ps2_current = self.read_value(root, entry, "/sample/ps2_current")
        self.sample_ps2_voltage = self.read_value(root, entry, "/sample/ps2_voltage")
        self.sample_ps3_current = self.read_value(root, entry, "/sample/ps3_current")
        self.sample_ps3_voltage = self.read_value(root, entry, "/sample/ps3_voltage")

        # Magnet
        self.sample_field_actual = self.read_value(root, entry, "/sample/field_actual")

        # Shear
        self.sample_shearrate_actual = self.read_value(
            root, entry, "/sample/shearrate_actual"
        )

        # Read Detector Data, Monitor and Time Slices
        self.detector_detsum = self.read_value(root, entry, "/detector/detsum")
        self.detector_detrate = self.read_value(root, entry, "/detector/detrate")

        # Determine Measurement Mode, Single, Kinetic or TOF
        self.mode = int(self.read_value(root, entry, "/mode", default=0))
        if self.mode == 0:
            self.file_type = "mono"
        elif self.mode == 1:
            self.file_type = "tof"
            tof_params = self.read_value(root, entry, "/monitor1/time_of_flight")
            if tof_params is not None:
                self.tof_width = tof_params[0]
                self.tof_channels = tof_params[1]
                self.tof_delay = tof_params[2]
                self.pickups = self.read_value(root, entry, "/monitor1/nbpickup")
                self.tof_period = (self.tof_width * self.tof_channels) + self.tof_delay
                self.tof_mode = self.read_value(root, entry, "/tof/tof_mode", default=0)
        elif self.mode == 3:
            self.file_type = "kinetic"
            self.slices = self.read_value(root, entry, "/slices")
            if self.slices is not None:
                # time lapse of the first frame - added to below with subsequent frames
                self.frame_time = self.slices[0] / 2
                self.pickups = self.read_value(root, entry, "/nbrepaint")
        elif self.mode == 4:
            self.file_type = "tof_inelastic"
            tof_params = self.read_value(root, entry, "/monitor1/time_of_flight")
            if tof_params is not None:
                self.tof_width = tof_params[0]
                self.tof_channels = tof_params[1]
                self.tof_delay = tof_params[2]
                self.pickups = self.read_value(root, entry, "/monitor1/nbpickup")
                self.tof_period = (self.tof_width * self.tof_channels) + self.tof_delay
        else:
            self.file_type = "unknown"

        # Detector Distance & Motors

        det = self.read_array(
            root, entry, "/D11/Detector 1/data1", fallback_item="/D11/Detector 1/data"
        )
        det_size = det.shape
        if len(det_size) < 3:
            det_size[2] = 1
        if self.file_type == "mono":
            det = det.reshape([det_size[2], det_size[0], det_size[1]])
            self.n_frames = det_size[2]
            self.det1_data = det[0, :, :].transpose()
            self.nx1, self.ny1 = self.det1_data.shape
        if self.file_type == "kinetic":
            if len(det_size) == 4:
                det = det.reshape(det_size[0], det_size[2], det_size[3])
            else:
                det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.n_frames = det_size[0]
            self.det1_data = det[0, :, :].transpose()
            self.nx1, self.ny1 = self.det1_data.shape
            self.duration = self.duration / self.n_frames
        if self.file_type == "tof":
            det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.det1_data = det[0, :, :].transpose()
            self.nx1, self.ny1 = self.det1_data.shape
            self.n_frames = det_size[0]
            self.duration = self.duration / self.n_frames
            # read tof wavelengths
            try:
                self.tof_wavs1 = self.read_value(
                    root, entry, "/tof/tof_wavelength_detector1"
                )
                self.tof_res1 = np.ones(self.n_frames)
            except Exception:
                self.tof_wavs1 = np.arrange(self.tof_channels)
                self.tof_res1 = np.ones(self.tof_channels)
        if self.file_type == "tof_inelastic":
            det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.det1_data = det[0, :, :].transpose()
            self.nx1, self.ny1 = self.det1_data.shape
            self.n_frames = det_size[0]
            self.duration = self.duration / self.n_frames
            # read tof wavelengths
            try:
                self.tof_wavs1 = self.read_value(
                    root, entry, "/tof/tof_wavelength_detector1"
                )
                self.tof_res1 = np.ones(self.n_frames)
            except Exception:
                self.tof_wavs1 = np.ones(self.n_frames) * self.wavelength
                self.tof_res1 = np.ones(self.n_frames) * self.wavelength_res

        det = self.read_array(
            root, entry, "/D11/Detector 2/data2", fallback_item="/D11/Detector 2/data"
        )
        det_size = det.shape
        if len(det_size) < 3:
            det_size[2] = 1
        if self.file_type == "mono":
            det = det.reshape([det_size[2], det_size[0], det_size[1]])
            self.n_frames = det_size[2]
            self.det2_data = det[0, :, :].transpose()
            self.nx2, self.ny2 = self.det2_data.shape
        if self.file_type == "kinetic":
            if len(det_size) == 4:
                det = det.reshape(det_size[0], det_size[2], det_size[3])
            else:
                det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.n_frames = det_size[0]
            self.det2_data = det[0, :, :].transpose()
            self.nx2, self.ny2 = self.det2_data.shape
            self.duration = self.duration / self.n_frames
        if self.file_type == "tof":
            det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.det2_data = det[0, :, :].transpose()
            self.nx2, self.ny2 = self.det2_data.shape
            self.n_frames = det_size[0]
            self.duration = self.duration / self.n_frames
            # read tof wavelengths
            try:
                self.tof_wavs2 = self.read_value(
                    root, entry, "/tof/tof_wavelength_detector1"
                )
                self.tof_res2 = np.ones(self.n_frames)
            except Exception:
                self.tof_wavs2 = np.arrange(self.tof_channels)
                self.tof_res2 = np.ones(self.tof_channels)
        if self.file_type == "tof_inelastic":
            det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.det2_data = det[0, :, :].transpose()
            self.nx2, self.ny2 = self.det2_data.shape
            self.n_frames = det_size[0]
            self.duration = self.duration / self.n_frames
            # read tof wavelengths
            try:
                self.tof_wavs2 = self.read_value(
                    root, entry, "/tof/tof_wavelength_detector1"
                )
                self.tof_res2 = np.ones(self.n_frames)
            except Exception:
                self.tof_wavs2 = np.ones(self.n_frames) * self.wavelength
                self.tof_res2 = np.ones(self.n_frames) * self.wavelength_res

        det = self.read_array(
            root, entry, "/D11/Detector 3/data3", fallback_item="/D11/Detector 3/data"
        )
        det_size = det.shape
        if len(det_size) < 3:
            det_size[2] = 1
        if self.file_type == "mono":
            det = det.reshape([det_size[2], det_size[0], det_size[1]])
            self.n_frames = det_size[2]
            self.det3_data = det[0, :, :].transpose()
            self.nx3, self.ny3 = self.det3_data.shape
        if self.file_type == "kinetic":
            if len(det_size) == 4:
                det = det.reshape(det_size[0], det_size[2], det_size[3])
            else:
                det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.n_frames = det_size[0]
            self.det3_data = det[0, :, :].transpose()
            self.nx3, self.ny3 = self.det3_data.shape
            self.duration = self.duration / self.n_frames
        if self.file_type == "tof":
            det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.det3_data = det[0, :, :].transpose()
            self.nx3, self.ny3 = self.det3_data.shape
            self.n_frames = det_size[0]
            self.duration = self.duration / self.n_frames
            # read tof wavelengths
            try:
                self.tof_wavs3 = self.read_value(
                    root, entry, "/tof/tof_wavelength_detector1"
                )
                self.tof_res3 = np.ones(self.n_frames)
            except Exception:
                self.tof_wavs3 = np.arrange(self.tof_channels)
                self.tof_res3 = np.ones(self.tof_channels)
        if self.file_type == "tof_inelastic":
            det = det.reshape(det_size[0], det_size[1], det_size[2])
            self.det3_data = det[0, :, :].transpose()
            self.nx3, self.ny3 = self.det3_data.shape
            self.n_frames = det_size[0]
            self.duration = self.duration / self.n_frames
            # read tof wavelengths
            try:
                self.tof_wavs3 = self.read_value(
                    root, entry, "/tof/tof_wavelength_detector1"
                )
                self.tof_res3 = np.ones(self.n_frames)
            except Exception:
                self.tof_wavs3 = np.ones(self.n_frames) * self.wavelength
                self.tof_res3 = np.ones(self.n_frames) * self.wavelength_res
        # Read Monitor Data - all frames
        self.monitor1_data = self.read_array(root, entry, "/monitor1/data", default=1)

    def read_value(self, root, entry, item, fallback_item=None, default=None):
        try:
            temp = root[entry + item]
            value = temp[0]
        except Exception:
            if fallback_item is not None:
                try:
                    temp = root[entry + fallback_item]
                    value = temp[0]
                except Exception:
                    # print("Bad file - Bad Fallback", item, fallback_item)
                    return default
            else:
                # print("Bad file - default", item, default)
                return default
        return value

    def read_array(self, root, entry, item, fallback_item=None, default=None):
        try:
            temp = root[entry + item]
            value = temp[()]
        except Exception:
            if fallback_item is not None:
                try:
                    temp = root[entry + fallback_item]
                    value = temp[()]
                except Exception:
                    # print("Bad file - Bad Fallback", item, fallback_item)
                    return default
            else:
                # print("Bad file - default", item, default)
                return default
        return value



from vtkmodules.vtkCommonCore import vtkLookupTable
from vtkmodules.vtkCommonDataModel import vtkImageData
from vtkmodules.numpy_interface.dataset_adapter import numpyTovtkDataArray as np2da
from vtkmodules.vtkRenderingCore import (
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
    vtkDataSetMapper,
    vtkPolyDataMapper,
    vtkActor,
    vtkDiscretizableColorTransferFunction,
)
from vtkmodules.vtkFiltersCore import (
    vtkCellDataToPointData,
    vtkContourFilter,
)
from vtkmodules.vtkRenderingAnnotation import (
    vtkScalarBarActor,
    vtkCubeAxesActor,
)

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import vtkInteractorStyleImage  # noqa
import vtkmodules.vtkRenderingOpenGL2  # noqa

from trame.decorators import TrameApp, change


def to_linear_value(v, data_range):
    return v * (data_range[1] - data_range[0]) + data_range[0]


def spectral(ctf, data_range):
    ctf.RemoveAllPoints()

    ctf.SetColorSpaceToRGB()
    ctf.SetScaleToLinear()

    ctf.SetNanColor(0.0, 0.0, 0.0)

    ctf.AddRGBPoint(
        to_linear_value(0 / (11 - 1), data_range),
        158.0 / 255.0,
        1.0 / 255.0,
        66.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(1 / (11 - 1), data_range),
        213.0 / 255.0,
        62.0 / 255.0,
        79.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(2 / (11 - 1), data_range),
        244.0 / 255.0,
        109.0 / 255.0,
        67.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(3 / (11 - 1), data_range),
        253.0 / 255.0,
        174.0 / 255.0,
        97.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(4 / (11 - 1), data_range),
        254.0 / 255.0,
        224.0 / 255.0,
        139.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(5 / (11 - 1), data_range),
        255.0 / 255.0,
        255.0 / 255.0,
        191.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(6 / (11 - 1), data_range),
        230.0 / 255.0,
        245.0 / 255.0,
        152.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(7 / (11 - 1), data_range),
        171.0 / 255.0,
        221.0 / 255.0,
        164.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(8 / (11 - 1), data_range),
        102.0 / 255.0,
        194.0 / 255.0,
        165.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(9 / (11 - 1), data_range),
        50.0 / 255.0,
        136.0 / 255.0,
        189.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(10 / (11 - 1), data_range),
        94.0 / 255.0,
        79.0 / 255.0,
        162.0 / 255.0,
    )
    ctf.SetNumberOfValues(11)
    ctf.DiscretizeOff()

    ctf.Build()

    return ctf


def rdbu(ctf, data_range):
    ctf.RemoveAllPoints()

    ctf.SetColorSpaceToRGB()
    ctf.SetScaleToLinear()

    ctf.SetNanColor(0.0, 0.0, 0.0)

    ctf.AddRGBPoint(
        to_linear_value(0 / (11 - 1), data_range),
        103.0 / 255.0,
        0.0 / 255.0,
        31.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(1 / (11 - 1), data_range),
        178.0 / 255.0,
        24.0 / 255.0,
        43.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(2 / (11 - 1), data_range),
        214.0 / 255.0,
        96.0 / 255.0,
        77.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(3 / (11 - 1), data_range),
        244.0 / 255.0,
        165.0 / 255.0,
        130.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(4 / (11 - 1), data_range),
        253.0 / 255.0,
        219.0 / 255.0,
        199.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(5 / (11 - 1), data_range),
        247.0 / 255.0,
        247.0 / 255.0,
        247.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(6 / (11 - 1), data_range),
        209.0 / 255.0,
        229.0 / 255.0,
        240.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(7 / (11 - 1), data_range),
        146.0 / 255.0,
        197.0 / 255.0,
        222.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(8 / (11 - 1), data_range),
        67.0 / 255.0,
        147.0 / 255.0,
        195.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(9 / (11 - 1), data_range),
        33.0 / 255.0,
        102.0 / 255.0,
        172.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(10 / (11 - 1), data_range),
        5.0 / 255.0,
        48.0 / 255.0,
        97.0 / 255.0,
    )
    ctf.SetNumberOfValues(11)
    ctf.DiscretizeOff()

    ctf.Build()

    return ctf


def gray(ctf, data_range):
    ctf.RemoveAllPoints()

    ctf.SetColorSpaceToRGB()
    ctf.SetScaleToLinear()

    ctf.SetNanColor(0.0, 0.5, 1.0)

    ctf.AddRGBPoint(
        to_linear_value(0 / (9 - 1), data_range),
        255.0 / 255.0,
        255.0 / 255.0,
        255.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(1 / (9 - 1), data_range),
        240.0 / 255.0,
        240.0 / 255.0,
        240.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(2 / (9 - 1), data_range),
        217.0 / 255.0,
        217.0 / 255.0,
        217.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(3 / (9 - 1), data_range),
        189.0 / 255.0,
        189.0 / 255.0,
        189.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(4 / (9 - 1), data_range),
        150.0 / 255.0,
        150.0 / 255.0,
        150.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(5 / (9 - 1), data_range),
        115.0 / 255.0,
        115.0 / 255.0,
        115.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(6 / (9 - 1), data_range),
        82.0 / 255.0,
        82.0 / 255.0,
        82.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(7 / (9 - 1), data_range),
        37.0 / 255.0,
        37.0 / 255.0,
        37.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(8 / (9 - 1), data_range), 0.0 / 255.0, 0.0 / 255.0, 0.0 / 255.0
    )
    ctf.SetNumberOfValues(9)
    ctf.DiscretizeOff()

    ctf.Build()

    return ctf


def ylgn(ctf, data_range):
    ctf.RemoveAllPoints()

    ctf.SetColorSpaceToRGB()
    ctf.SetScaleToLinear()

    ctf.SetNanColor(1.0, 0.5, 0.0)

    ctf.AddRGBPoint(
        to_linear_value(0 / (9 - 1), data_range),
        255.0 / 255.0,
        255.0 / 255.0,
        229.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(1 / (9 - 1), data_range),
        247.0 / 255.0,
        252.0 / 255.0,
        185.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(2 / (9 - 1), data_range),
        217.0 / 255.0,
        240.0 / 255.0,
        163.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(3 / (9 - 1), data_range),
        173.0 / 255.0,
        221.0 / 255.0,
        142.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(4 / (9 - 1), data_range),
        120.0 / 255.0,
        198.0 / 255.0,
        121.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(5 / (9 - 1), data_range),
        65.0 / 255.0,
        171.0 / 255.0,
        93.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(6 / (9 - 1), data_range),
        35.0 / 255.0,
        132.0 / 255.0,
        67.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(7 / (9 - 1), data_range),
        0.0 / 255.0,
        104.0 / 255.0,
        55.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(8 / (9 - 1), data_range),
        0.0 / 255.0,
        69.0 / 255.0,
        41.0 / 255.0,
    )
    ctf.SetNumberOfValues(9)
    ctf.DiscretizeOff()

    ctf.Build()

    return ctf


def ylorbr(ctf, data_range):
    ctf.RemoveAllPoints()

    ctf.SetColorSpaceToRGB()
    ctf.SetScaleToLinear()

    ctf.SetNanColor(0.0, 0.5, 1.0)

    ctf.AddRGBPoint(
        to_linear_value(0 / (9 - 1), data_range),
        255.0 / 255.0,
        255.0 / 255.0,
        229.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(1 / (9 - 1), data_range),
        255.0 / 255.0,
        247.0 / 255.0,
        188.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(2 / (9 - 1), data_range),
        254.0 / 255.0,
        227.0 / 255.0,
        145.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(3 / (9 - 1), data_range),
        254.0 / 255.0,
        196.0 / 255.0,
        79.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(4 / (9 - 1), data_range),
        254.0 / 255.0,
        153.0 / 255.0,
        41.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(5 / (9 - 1), data_range),
        236.0 / 255.0,
        112.0 / 255.0,
        20.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(6 / (9 - 1), data_range),
        204.0 / 255.0,
        76.0 / 255.0,
        2.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(7 / (9 - 1), data_range),
        153.0 / 255.0,
        52.0 / 255.0,
        4.0 / 255.0,
    )
    ctf.AddRGBPoint(
        to_linear_value(8 / (9 - 1), data_range),
        102.0 / 255.0,
        37.0 / 255.0,
        6.0 / 255.0,
    )
    ctf.SetNumberOfValues(9)
    ctf.DiscretizeOff()

    ctf.Build()

    return ctf


LUTS = {
    "spectral": spectral,
    "rdbu": rdbu,
    "gray": gray,
    "ylgn": ylgn,
    "ylorbr": ylorbr,
}


@TrameApp()
class VTKDetectorView:
    def __init__(self, server, show_scalar_bar=False):
        self.server = server

        self.color_range = [0, 1]
        self.lut = vtkDiscretizableColorTransferFunction()

        self.renderer = vtkRenderer()
        self.renderer.SetBackground(1.0, 1.0, 1.0)
        self.render_window = vtkRenderWindow()
        self.render_window.AddRenderer(self.renderer)
        self.render_window.OffScreenRenderingOn()

        self.render_window_interactor = vtkRenderWindowInteractor()
        self.render_window_interactor.SetRenderWindow(self.render_window)
        self.render_window_interactor.SetInteractorStyle(vtkInteractorStyleImage())
        # self.render_window_interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        # Heatmap
        self.mapper = vtkDataSetMapper()
        self.mapper.SetLookupTable(self.lut)
        self.actor = vtkActor()
        self.actor.SetVisibility(0)
        self.actor.SetMapper(self.mapper)
        self.renderer.AddActor(self.actor)

        lut = vtkLookupTable()
        lut.SetRange(0.0, 1.0)
        lut.SetNanColor(1.0, 0.0, 0.0, 0.0)
        lut.SetRampToLinear()
        lut.SetHueRange(0.0, 0.0)
        lut.SetSaturationRange(0.0, 0.0)
        lut.SetValueRange(0.0, 1.0)

        self.mask_mapper = vtkDataSetMapper()
        self.mask_mapper.SetScalarVisibility(True)
        self.mask_mapper.SetLookupTable(lut)
        self.mask_actor = vtkActor()
        self.mask_actor.SetVisibility(0)
        self.mask_actor.SetMapper(self.mask_mapper)
        self.renderer.AddActor(self.mask_actor)
        # Contour
        self.c2p = vtkCellDataToPointData()
        self.contour = vtkContourFilter()
        self.contour_mapper = vtkPolyDataMapper()
        self.contour_mapper.SetLookupTable(self.lut)
        self.contour_mapper.SetScalarVisibility(True)
        self.contour_actor = vtkActor()
        self.contour_actor.SetVisibility(0)
        self.contour_actor.SetMapper(self.contour_mapper)
        self.contour_actor.GetProperty().SetLineWidth(2)
        self.contour_actor.GetProperty().SetColor(0, 0, 0)
        self.renderer.AddActor(self.contour_actor)

        self.cubeaxes = vtkCubeAxesActor()
        self.cubeaxes.SetCamera(self.renderer.GetActiveCamera())
        font_size = 9
        self.cubeaxes.SetScreenSize(font_size / 12 * 10.0)
        self.cubeaxes.GetXAxesLinesProperty().SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetTitleTextProperty(0).SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetLabelTextProperty(0).SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetYAxesLinesProperty().SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetTitleTextProperty(1).SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetLabelTextProperty(1).SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetZAxesLinesProperty().SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetTitleTextProperty(2).SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.GetLabelTextProperty(2).SetColor(0.0, 0.0, 0.0)
        self.cubeaxes.SetXTitle("")
        self.cubeaxes.SetYTitle("")
        self.cubeaxes.SetZTitle("")
        self.cubeaxes.SetTickLocation(1)
        self.renderer.AddActor(self.cubeaxes)

        self.image_data = None
        self.html_update = None
        self.html_reset_camera = None

        self.scalar_bar = None
        if show_scalar_bar:
            self.scalar_bar = vtkScalarBarActor()
            self.scalar_bar.SetLookupTable(self.lut)
            self.scalar_bar.SetOrientationToHorizontal()
            self.scalar_bar.SetPosition(0.1, 0.01)
            self.scalar_bar.SetPosition2(0.8, 0.1)
            self.scalar_bar.GetLabelTextProperty().SetColor(0.0, 0.0, 0.0)
            self.scalar_bar.GetLabelTextProperty().SetFontSize(font_size)
            self.scalar_bar.GetLabelTextProperty().BoldOff()
            self.scalar_bar.GetLabelTextProperty().ItalicOff()
            self.scalar_bar.GetLabelTextProperty().ShadowOff()
            self.scalar_bar.SetUnconstrainedFontSize(True)
            self.renderer.AddActor(self.scalar_bar)

    @change("selectedColor")
    def on_preset_change(self, **kwargs):
        self.update_lut()
        self.html_update()

    def set_html_view(self, view):
        self.html_update = view.update
        self.html_reset_camera = view.reset_camera

        if self.image_data:
            self.html_reset_camera()
            self.html_update()

    def update_lut(self):
        LUTS[self.server.state.selectedColor](self.lut, self.color_range)

    def set_data_minmax(self, data_min, data_max):
        self.color_range = [data_min, data_max]

    def update_data(self, np_array, mask, x, y):
        state = self.server.state

        self.image_data = vtkImageData()
        self.image_data.SetSpacing(x, y, 1)

        self.image_data.SetDimensions(np_array.shape[1] + 1, np_array.shape[0] + 1, 1)

        vtk_array = np2da(np_array.flatten(), name="scalar")
        self.image_data.GetCellData().SetScalars(vtk_array)

        self.mask_data = vtkImageData()
        self.mask_data.SetSpacing(x, y, 1)

        self.mask_data.SetDimensions(mask.shape[1] + 1, mask.shape[0] + 1, 1)

        vtk_mask = np2da(mask.flatten(), name="scalar")
        self.mask_data.GetCellData().SetScalars(vtk_mask)

        # self.color_range = vtk_array.GetRange()
        self.update_lut()

        # Heatmap
        if state.selectedRepresentation == "Heatmap":
            self.mapper.SetScalarRange(self.color_range)
            self.mapper.SetInputData(self.image_data)
            self.actor.SetVisibility(1)
            if state.show_mask:
                self.mask_mapper.SetInputData(self.mask_data)
                self.mask_actor.SetVisibility(1)
            else:
                self.mask_actor.SetVisibility(0)
            self.contour_actor.SetVisibility(0)
        # Contours
        elif state.selectedRepresentation == "Contours":
            self.c2p.SetInputData(self.image_data)
            self.c2p.Update()
            self.contour.SetInputData(self.c2p.GetOutput())
            self.contour.SetArrayComponent(0)
            self.contour.GenerateValues(10, self.color_range)
            self.contour.Update()
            self.contour_mapper.SetInputConnection(self.contour.GetOutputPort())
            self.contour_mapper.SetScalarVisibility(True)
            self.contour_mapper.SetScalarRange(self.color_range)
            self.contour_actor.SetVisibility(1)
            self.actor.SetVisibility(0)
            self.mask_actor.SetVisibility(0)
        else:
            self.mapper.SetScalarRange(self.color_range)
            self.mapper.SetInputData(self.image_data)
            self.actor.SetVisibility(1)
            if state.show_mask:
                self.mask_mapper.SetInputData(self.mask_data)
                self.mask_actor.SetVisibility(1)
            else:
                self.mask_actor.SetVisibility(0)
            self.c2p.SetInputData(self.image_data)
            self.c2p.Update()
            self.contour.SetInputData(self.c2p.GetOutput())
            self.contour.SetArrayComponent(0)
            self.contour.GenerateValues(10, self.color_range)
            self.contour.Update()
            self.contour_mapper.SetInputConnection(self.contour.GetOutputPort())
            self.contour_mapper.SetScalarRange(self.color_range)
            self.contour_mapper.SetScalarVisibility(False)
            self.contour_actor.SetVisibility(1)

        if state.show_mask and state.selectedRepresentation != "Contours":
            self.cubeaxes.SetBounds(self.mask_data.GetBounds())
        else:
            self.cubeaxes.SetBounds(self.image_data.GetBounds())

        if self.html_reset_camera:
            self.html_reset_camera()
            self.html_update()
